.loading-frame {
  min-height: 200px;
}

.architecture button:not(:hover) {
  background: transparent;
}

.app_icon {
  -gtk-icon-size: 128px;
  margin-bottom: 24px;
}

.drag-overlay-status-page {
  background-color: alpha(@accent_bg_color, 0.5);
  color: @accent_fg_color;
}

.error-message {
  margin: 12px;
}
