#!/usr/bin/env bash

BUILD_DIR="translation-build/"
if [ -d "$BUILD_DIR" ]; then
	rm -r translation-build
fi

meson setup translation-build
meson compile -C translation-build impression-pot

rm -r translation-build
